

         /*********************************************************/
         /*                                                       */
         /*                 COPYRIGHT (C) 1988 BY                 */
         /* CAD Group of Dipartimento di Automatica e Informatica */
         /*             Politecnico di Torino, Italy.             */
         /*                  ALL RIGHTS RESERVED.                 */
         /*                                                       */
         /* This software is furnished under a license and may be */
         /* used and copied  only in accordance with the terms of */
         /* such license. Unauthorized copies will be persecuted. */
         /*                                                       */
         /*********************************************************/





/******************************************************************************

    Filename:         cwsl.c

    Author:           Pier Luca Montessoro

    Creation date:    1 August 1988

    Contents:         WSL compile for Dr. Creator simulator


                          *** DOCUMENTATION ***

*******************************

 Source file format:

Header:

   $input
   in1
   in2
   ...

   $output
   out1
   out2
   ...

   $clock   (it works only for edge sensitive elements: clock suppression)
   inp_clock1 period rise_phase fall_phase
   inp_clock2 period rise_phase fall_phase
   ...


   $begin

Statements:

   start_time $period
              input
              0 value
              t1 value
              ...
              tn $endperiod

   time       $strobe

   time       $strobeperiod  time

   time       input  value

   time       $end

********************************************************************************

 Compiled file format:

Header:

   # observation points (for strobed fault detection and checks)

   1st obs. point index
   2nd obs. point index
   ...

   # clock inputs
   1st clock input index       period       rise phase       fall phase
   2nd clock input index       period       rise phase       fall phase
   ...

Statements:

   END code
       
   NEXT code           next time

   FORCE code          index         value string

   MOD_FORCE code      index         value string

   FREE code           index

   STROBE code         index

*******************************


 MAIN

      is the main function.


 DOC_COMPILE

      is the compiler for Dr. Creator.


 MOZART_COMPILE

      is the compiler (translator to a new source form) for MOZART.


 HILO_COMPILE

      is the compiler (translator to a new source form) for HILO.


 VERILOG_COMPILE

      is the compiler (translator to a new source form) for VERILOG.


 LOAD_PERIOD

      loads a periodical signal.


 NEXT_PERIOD

      searches the next activation of periodic signals.


 DOC_WRITE_NEXT_PERIOD

      writes in Dr. Creator format the next activation of periodical signals.


 MOZART_WRITE_NEXT_PERIOD

      writes in MOZART format the next activation of periodical signals.


 HILO_WRITE_NEXT_PERIOD

      writes in HILO format the next activation of periodical signals.


 VERILOG_WRITE_NEXT_PERIOD

      writes in VERILOG format the next activation of periodical signals.


 LOAD_STROBE_PERIOD

      loads a periodical strobe signal.


 GOOD_BYE

      is the exit function.


 GET_MEMORY

      allocs the memory needed for the topological data structures.


 LOAD_SYM_TAB

      loads the symbol table of the main unit.


 REDUCE_SYM_TAB

      removes useless entries form symbol table after hen inputs and autputs 
 have been recognized.


 EL_INDEX
 ST_INDEX

      return the index and the position in the symbol table of a named element.


 UPPERCASE

      is trivial.


 US_CHANGE

      replaces '_' with 'X' in names (useful when translating to hilo).


 IS_INTEGER
 IS_VALUE

      perform some checks.


 ******************************************************************************/



#include "env.h"
#include "util.h"


#define  MAX_FILE_NAME_LENGTH  64
#define  MAX_NAME_LENGTH       32
#define  MAX_MESSAGE_LENGTH   132
#define  MAX_DATA_SIZE        112
#define  BUFFER_SIZE           64

#define  MAX_CLOCK_SIGNALS     16
#define  MAX_PERIOD_LENGTH     64

#define  SPAT_EXTENSION     "dws"
#define  DOC_EXTENSION      "dwe"
#define  HILO_EXTENSION     "hws"
#define  MOZART_EXTENSION   "tes"
#define  VERILOG_EXTENSION  "vin"
#define  SYMTAB_EXTENSION   "smt"


/* signal structure */

typedef struct Signal
        {
           unsigned long  time;
           char           value;
           char           unused_1;
           short          unused_2;

        }  SIGNAL;


/* periodical signal structure */

typedef struct Periodic_Signal
        {
           struct Periodic_Signal  *next_signal;
           unsigned long           st_index;
           unsigned long           time_offset;
           SIGNAL                  *next_value;
           SIGNAL                  table[MAX_PERIOD_LENGTH];

        }  PERIODIC_SIGNAL;


/* symbol table element structure definition */

typedef struct Sym_Tab
        {
           unsigned char   is_output;
           unsigned char   is_input;
           unsigned char   is_clock;
           char            last_value;
           unsigned long   last_time;
           unsigned long   index;
           char            name [MAX_NAME_LENGTH];

        }  SYM_TAB;


/* clock table element definition */

typedef struct Clock_Tab
        {
           unsigned long   index;
           unsigned long   period;
           unsigned long   rise_phase;
           unsigned long   fall_phase;

        }  CLOCK_TAB;


/* symbol table */

SYM_TAB                *sym_tab;
unsigned long          sym_tab_dim;
unsigned char          us_converted;


/* clock table */

CLOCK_TAB              clock_tab[MAX_CLOCK_SIGNALS];
unsigned long          clock_signals;


/* periodic signals list head */

PERIODIC_SIGNAL        *p_list_head = NULL;


/* auxiliary variable */

PERIODIC_SIGNAL        *next_p_act_signal = NULL;


/* last strobe time */

unsigned long          last_strobe_time = MAX_UNSIGNED;


/* last_time for verilog compiler */

unsigned long          v_last_time = 0;


/* input WSL file */

FILE                   *wsl_file;


/* output file */

FILE                   *out_file;


/******************************************************************************

    Function name:    main

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          main

 ******************************************************************************/


main (argc, argv)

int   argc;
char  *argv[];

{

   /* local variables */

   char                  wsl_file_name [MAX_FILE_NAME_LENGTH];
   char                  out_file_name [MAX_FILE_NAME_LENGTH];


   /* beginning of executable part */

   if (argc != 4)
   {
      printf 
 ("Wrong number of arguments (1st: wavesfile, 2nd: netfile, 3rd: d/h/m/v).\n");
      good_bye (FALSE, TRUE);
   }


   /* open the source and the output files */

   strcpy (wsl_file_name, argv[1]);
   strcat (wsl_file_name, ".");
   strcat (wsl_file_name, SPAT_EXTENSION);

   strcpy (out_file_name, argv[1]);
   strcat (out_file_name, ".");

   switch (*(argv[3]))
   {
      case 'd':
      case 'D':
         strcat (out_file_name, DOC_EXTENSION);
         break;

      case 'm':
      case 'M':
         strcat (out_file_name, MOZART_EXTENSION);
         break;

      case 'h':
      case 'H':
         strcat (out_file_name, HILO_EXTENSION);
         break;

      case 'v':
      case 'V':
         strcat (out_file_name, VERILOG_EXTENSION);
         break;

      default:
         printf ("I cannot understand the second argument.\n");
         good_bye (FALSE, TRUE);
   }

   if ((wsl_file = fopen (wsl_file_name, "r")) == NULL)
   {
      printf ("I cannot open the source file.\n");
      good_bye (TRUE, TRUE);
   }

   if ((out_file = fopen (out_file_name, "w")) == NULL)
   {
      printf ("I cannot open the output file.\n");
      good_bye (TRUE, TRUE);
   }

   load_sym_tab (argv[2], *(argv[3]));

   switch (*(argv[3]))
   {
      case 'd':
      case 'D':
         doc_compile();
         break;

      case 'm':
      case 'M':
         mozart_compile();
         break;

      case 'h':
      case 'H':
         hilo_compile(argv[1]);
         break;

      case 'v':
      case 'V':
         verilog_compile(argv[1]);
         break;

      default:
         printf ("I cannot understand the second argument.\n");
         good_bye (FALSE, TRUE);
   }

   printf ("Output file %s written.\n", out_file_name);

   good_bye (TRUE, FALSE);

}  /* end of main */


/******************************************************************************

    Function name:    doc_compile

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          to compile the wsl file

 ******************************************************************************/


doc_compile ()

{

   /* called function */

   unsigned long  next_period();
   char           *get_memory();
   unsigned long  el_index();
   unsigned long  st_index();
   char           *uppercase();
   unsigned char  is_integer();
   unsigned char  is_value();


   /* local variables */

   char           buffer [BUFFER_SIZE];
   unsigned long  i;
   unsigned long  cnt;
   unsigned long  curr_time;
   unsigned long  next_time;
   unsigned long  period_next_time;
   unsigned long  tmp_index;
   unsigned long  tmp_st_index;
   unsigned char  end_of_pat;


   /* beginning of executable part */

   /* read the header */

   /* inputs (list may be empty) */

   fscanf (wsl_file, "%s", buffer);

   if (strcmp (uppercase (buffer), "$INPUT"))
   {
      printf ("I can't find the $INPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   fscanf (wsl_file, "%s", buffer);

   while (strcmp (uppercase (buffer), "$OUTPUT"))
   {
      sym_tab [st_index (buffer)].is_input = TRUE;
      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("I can't find the $OUTPUT statement.\n");
         good_bye (TRUE, TRUE);
      }
   }


   /* outputs (list may be empty) */

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $OUTPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   while (strcmp (uppercase (buffer), "$CLOCK"))
   {
      sym_tab [st_index (buffer)].is_output = TRUE;
      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("$CLOCK statement not found.\n");
         good_bye (TRUE, TRUE);
      }
   }


   /* clock signals (list may be empty) */

   clock_signals = 0;

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("$BEGIN statement not found.\n");
      good_bye (TRUE, TRUE);
   }
   while (strcmp (uppercase (buffer), "$BEGIN"))
   {
      if (clock_signals >= MAX_CLOCK_SIGNALS)
      {
         printf ("Too many clock signals.\n");
         good_bye (TRUE, TRUE);
      }

      fscanf (wsl_file, "%ld %ld %ld", 
              &clock_tab[clock_signals].period, 
              &clock_tab[clock_signals].rise_phase, 
              &clock_tab[clock_signals].fall_phase);
      tmp_st_index = st_index (buffer);
      clock_tab[clock_signals].index = sym_tab [tmp_st_index].index;
      sym_tab [tmp_st_index].is_clock = TRUE;
      clock_signals++;
      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("$BEGIN statement not found.\n");
         good_bye (TRUE, TRUE);
      }
   }


   /* reduce the size of symbol table */

   reduce_sym_tab();


   /* write the output declaration */

   for (i = 0, cnt = 0; i < sym_tab_dim; i++)
   {
      if (sym_tab[i].is_output) cnt++;
   }

   fprintf (out_file, "%lu\n", cnt);

   for (i = 0, cnt = 0; i < sym_tab_dim; i++)
   {
      if (sym_tab[i].is_output)
      {
         fprintf (out_file, "%lu\n", sym_tab[i].index);
      }
   }


   /* write the clock signals declaration */

   fprintf (out_file, "%lu\n", clock_signals);

   for (i = 0; i < clock_signals; i++)
   {
      fprintf (out_file, "%lu %lu %lu %lu\n", 
              clock_tab[i].index, 
              clock_tab[i].period, 
              clock_tab[i].rise_phase, 
              clock_tab[i].fall_phase);
   }


   /* patterns */

   curr_time = 0;
   end_of_pat = FALSE;

   while ((fscanf (wsl_file, "%s", buffer) != EOF) && (!end_of_pat))
   {
      if (!is_integer (buffer))
      {
         printf ("Found %s when time was expected. Last time: %lu.\n",
                                                        buffer, curr_time);
         good_bye (TRUE, TRUE);
      }
      sscanf (buffer, "%ldu", &next_time);

      if (next_time != curr_time)
      {
         if (next_time < curr_time)
         {
            printf ("Time cannot proceed backward! Last time: %lu.\n", 
                                                               curr_time);
            good_bye (TRUE, TRUE);
         }

         while ((period_next_time = next_period()) < next_time)
         {
            if (curr_time != period_next_time)
            {
               fprintf (out_file, "%d %lu\n", WSL_NEXT, period_next_time);
               curr_time = period_next_time;
            }
            doc_write_next_period ();
         }
         curr_time = next_time;
         fprintf (out_file, "%d %lu\n", WSL_NEXT, next_time);
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
         good_bye (TRUE, TRUE);
      }

      if (buffer[0] != '$')
      {
         tmp_index = st_index (uppercase (buffer));
         if (sym_tab [tmp_index].is_clock) 
         {
            printf 
            ("Element %s is a clock input and cannot be forced at time %lu.\n",
             buffer, curr_time);
            good_bye (TRUE, TRUE);
         }
         if (sym_tab [tmp_index].last_time == curr_time)
         {
            printf ("Element %s forced more than once at time %lu.\n",
                    buffer, curr_time);
            good_bye (TRUE, TRUE);
         }

         if (fscanf (wsl_file, "%s", buffer) == EOF)
         {
            printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }
         if (!is_value (uppercase(buffer)))
         {
            printf ("Value expected. Last time: %lu\n", curr_time);
            good_bye (TRUE, TRUE);
         }

         /* filter unnecessary schedulations */
         if (buffer[0] != sym_tab[tmp_index].last_value)
         {
            if (sym_tab[tmp_index].last_time == MAX_UNSIGNED)
            {
               fprintf (out_file, "%d ", WSL_FORCE);
            }
            else
            {
               fprintf (out_file, "%d ", WSL_MOD_FORCE);
            }
            sym_tab[tmp_index].last_value = buffer[0];
            fprintf (out_file, "%lu %s\n", sym_tab[tmp_index].index, buffer);
            sym_tab[tmp_index].last_time = curr_time;
         }
      }
      else
      {
         if (!strcmp (uppercase (buffer), "$STROBE"))
         {
            fprintf (out_file, "%d\n", WSL_STROBE);
            last_strobe_time = curr_time;
         }
         else
         if (!strcmp (buffer, "$PERIOD"))
         {
            load_period (curr_time, 'D');
         }
         else
         if (!strcmp (buffer, "$STROBEPERIOD"))
         {
            load_strobe_period (curr_time);
         }
         else
         if (!strcmp (buffer, "$END"))
         {
            end_of_pat = TRUE;
         }
         else
         {
            printf ("%s statement is unknown.\n", buffer);
            good_bye (TRUE, TRUE);
         }
      }
   }


   if (!end_of_pat)
   {
      printf ("Warning: $END statement not found.\n");
   }

   fprintf (out_file, "%d\n", WSL_END);

   return;

}  /* end of doc_compile */


/******************************************************************************

    Function name:    mozart_compile

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          to compile the wsl file

 ******************************************************************************/


mozart_compile ()

{

   /* called function */

   unsigned long  next_period();
   char           *get_memory();
   unsigned long  el_index();
   unsigned long  st_index();
   char           *uppercase();
   unsigned char  is_integer();
   unsigned char  is_value();


   /* local variables */

   char           buffer [BUFFER_SIZE];
   char           value [BUFFER_SIZE];
   char           element [BUFFER_SIZE];
   unsigned long  i;
   unsigned long  cnt;
   unsigned long  curr_time;
   unsigned long  next_time;
   unsigned long  tmp_index;
   unsigned char  comma;
   unsigned char  end_of_pat;
   unsigned char  already_found_$strobeperiod;
   unsigned char  strobeperiod_found;
   unsigned long  s_start_time, s_time_period;


   /* beginning of executable part */

   /* read the header */

   /* inputs (list may NOT be empty) */

   fscanf (wsl_file, "%s", buffer);

   if (strcmp (uppercase (buffer), "$INPUT"))
   {
      printf ("I can't find the $INPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   cnt = 0;

   fscanf (wsl_file, "%s", buffer);

   while (strcmp (uppercase (buffer), "$OUTPUT"))
   {
      cnt++;
      sym_tab [st_index (buffer)].is_input = TRUE;

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("I can't find the $OUTPUT statement.\n");
         good_bye (TRUE, TRUE);
      }
   }

   if (cnt == 0)
   {
      printf ("No input declared.\n");
      good_bye (TRUE, TRUE);
   }


   /* outputs (list may NOT be empty) */

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $OUTPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   cnt = 0;

   while (strcmp (uppercase (buffer), "$CLOCK"))
   {
      sym_tab [st_index (buffer)].is_output = TRUE;
      cnt++;

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("$BEGIN statement not found.\n");
         good_bye (TRUE, TRUE);
      }
   }

   if (cnt == 0)
   {
      printf ("No output declared.\n");
      good_bye (TRUE, TRUE);
   }


   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $CLOCK statement.\n");
      good_bye (TRUE, TRUE);
   }

   if (strcmp (uppercase (buffer), "$BEGIN"))
   {
      printf ("I cannot translate the $CLOCK statement.\n");
      good_bye (TRUE, TRUE);
   }


   /* reduce the size of symbol table */

   reduce_sym_tab();


   /* search for the $strobeperiod declaration */

   {
      char           b [BUFFER_SIZE], f [BUFFER_SIZE];


      strobeperiod_found = FALSE;

      b [0] = '0';
      b [1] = '\0';

      while (fscanf (wsl_file, "%s", f) != EOF)
      {
         if (!strcmp (uppercase (f), "$STROBEPERIOD"))
         {
            if (!is_integer (b))
            {
               printf ("Wrong starting time for strobe period.\n");
               good_bye (TRUE, TRUE);
            }

            sscanf (b, "%ld", &s_start_time);

            if (fscanf (wsl_file, "%s", b) == EOF)
            {
               printf 
        ("Premature end of WSL file encountered. Strobe period at time: %lu.\n",
                                                                  s_start_time);
               good_bye (TRUE, TRUE);
            }

            sscanf (b, "%ld", &s_time_period);

            if (s_start_time >= s_time_period)
            {
               printf ("I cannot translate the strobe period command.\n");
               good_bye (TRUE, TRUE);
            }

            fprintf (out_file, "\ntime_unit 1ns;\ntest_step %luns;\n\n", 
                                s_time_period);

            strobeperiod_found = TRUE;

            break;
         }

         strcpy (b, f);
      }

      if (!strobeperiod_found)
      {
         fprintf (out_file, "\ntime_unit 1ns;\ntest_step 1ns;\n\n");
      }

      rewind (wsl_file);
      do 
      {
         (fscanf (wsl_file, "%s", buffer));
      }  while (strcmp (uppercase (buffer), "$BEGIN"));

   }


   /* write input list */

   fprintf (out_file, "in\n");

   comma = FALSE;

   for (cnt = 0; cnt < sym_tab_dim; cnt++)
   {
      if (sym_tab [cnt].is_input)
      {
         if (comma)
         {
            fprintf (out_file, ",\n   %s", sym_tab[cnt].name);
         }
         else
         {
            fprintf (out_file, "   %s", sym_tab[cnt].name);
            comma = TRUE;
         }
      }
   }

   fprintf (out_file, ";\n\n");


   /* write output list */

   fprintf (out_file, "out\n");

   comma = FALSE;

   for (cnt = 0; cnt < sym_tab_dim; cnt++)
   {
      if (sym_tab [cnt].is_output)
      {
         if (comma)
         {
            fprintf (out_file, ",\n   %s", sym_tab[cnt].name);
         }
         else
         {
            fprintf (out_file, "   %s", sym_tab[cnt].name);
            comma = TRUE;
         }
      }
   }

   fprintf (out_file, ";\n");


   /* strobe timeset, if declared */

   if (strobeperiod_found)
   {
      fprintf (out_file, "\ntimeset strobe %lu:\n", s_start_time);

      /* write (again, sigh!) output list */
      
      comma = FALSE;
      
      for (cnt = 0; cnt < sym_tab_dim; cnt++)
      {
         if (sym_tab [cnt].is_output)
         {
            if (comma)
            {
               fprintf (out_file, ",\n   %s", sym_tab[cnt].name);
            }
            else
            {
               fprintf (out_file, "   %s", sym_tab[cnt].name);
               comma = TRUE;
            }
         }
      }
      fprintf (out_file, ";\n");
   }


   /* patterns */

   fprintf (out_file, "\n\nbegin\n\n");

   curr_time = 0;
   end_of_pat = FALSE;
   already_found_$strobeperiod = FALSE;

   while ((fscanf (wsl_file, "%s", buffer) != EOF) && (!end_of_pat))
   {
      if (!is_integer (buffer))
      {
         printf ("Found %s when time was expected. Last time: %lu.\n", 
                                                        buffer, curr_time);
         good_bye (TRUE, TRUE);
      }
      sscanf (buffer, "%ld", &next_time);

      if (next_time != curr_time)
      {
         if (next_time < curr_time)
         {
            printf ("Time cannot proceed backward! Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }

         while ((curr_time = next_period()) < next_time)
         {
            mozart_write_next_period ();
         }
         curr_time = next_time;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
         good_bye (TRUE, TRUE);
      }

      if (buffer[0] != '$')
      {
         strcpy (element, buffer);

         if (!sym_tab [tmp_index = st_index (uppercase (element))].is_input)
         {
            printf ("%s has not been declared as input.\n", element);
            good_bye (TRUE, TRUE);
         }

         if (sym_tab [tmp_index].last_time == curr_time)
         {
            printf ("Element %s forced more than once at time %lu.\n",
                    element, curr_time);
            good_bye (TRUE, TRUE);
         }

         if (fscanf (wsl_file, "%s", value) == EOF)
         {
            printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }
         if (!is_value (uppercase(value)))
         {
            printf ("Value expected. Last time: %lu.\n", curr_time);
            good_bye (TRUE, TRUE);
         }

         /* filter unnecessary schedulations */
         if (value[0] != sym_tab[tmp_index].last_value)
         {
            fprintf (out_file, "#%lu : %s := %s;\n", curr_time, element, value);
            sym_tab[tmp_index].last_value = value[0];
            sym_tab[tmp_index].last_time = curr_time;
         }
      }
      else
      {
         if (!strcmp (uppercase (buffer), "$STROBE"))
         {
            printf
          ("Warning: $strobe command is ignored when translating to Mozart.\n");
         }
         else
         if (!strcmp (buffer, "$PERIOD"))
         {
            load_period (curr_time, 'H');
         }
         else
         if (!strcmp (buffer, "$STROBEPERIOD"))
         {
            if (already_found_$strobeperiod)
            {
               printf 
                  ("Warning: only first $strobeperiod command is meaningful\n");
               printf ("   when translating to Mozart.\n");
            }
            else
            {
               already_found_$strobeperiod = TRUE;
            }

            /* skip the time */
            fscanf (wsl_file, "%s", buffer);
         }
         else
         if (!strcmp (buffer, "$END"))
         {
             end_of_pat = TRUE;
         }
         else
         {
            printf ("%s statement is unknown.\n", buffer);
            good_bye (TRUE, TRUE);
         }
      }
   }


   if (!end_of_pat)
   {
      printf ("Warning: $END statement not found.\n");
   }

   fprintf (out_file, "#%lu : end;\n", curr_time);

   return;

}  /* end of mozart_compile */


/******************************************************************************

    Function name:    hilo_compile

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          to compile the wsl file

 ******************************************************************************/


hilo_compile (wave_name)

char  *wave_name;

{

   /* called function */

   unsigned long  next_period();
   char           *get_memory();
   unsigned long  el_index();
   unsigned long  st_index();
   char           *uppercase();
   unsigned char  is_integer();
   unsigned char  is_value();


   /* local variables */

   char           buffer [BUFFER_SIZE];
   char           value [BUFFER_SIZE];
   char           element [BUFFER_SIZE];
   unsigned long  i;
   unsigned long  cnt;
   unsigned long  curr_time;
   unsigned long  next_time;
   unsigned long  tmp_index;
   unsigned char  comma;
   unsigned char  end_of_pat;


   /* beginning of executable part */

   /* write the waveform name */

   fprintf (out_file, "waveform %spat\n\n", wave_name);

   /* read the header */

   /* inputs (list may NOT be empty) */

   fscanf (wsl_file, "%s", buffer);

   if (strcmp (uppercase (buffer), "$INPUT"))
   {
      printf ("I can't find the $INPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   fprintf (out_file, "stimulus\n", wave_name);

   comma = FALSE;
   cnt = 0;

   fscanf (wsl_file, "%s", buffer);

   while (strcmp (uppercase (buffer), "$OUTPUT"))
   {
      cnt++;
      us_change (buffer);
      sym_tab [st_index (buffer)].is_input = TRUE;

      if (comma)
      {
         fprintf (out_file, ",\n   %s", buffer);
      }
      else
      {
         fprintf (out_file, "   %s", buffer);
         comma = TRUE;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("I can't find the $OUTPUT statement.\n");
         good_bye (TRUE, TRUE);
      }
   }

   if (cnt == 0)
   {
      printf ("No input declared.\n");
      good_bye (TRUE, TRUE);
   }

   fprintf (out_file, " = x;\n\n");


   /* outputs (list may NOT be empty) */

   fprintf (out_file, "response\n");

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $OUTPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   comma = FALSE;
   cnt = 0;

   while (strcmp (uppercase (buffer), "$CLOCK"))
   {
      cnt++;
      us_change (buffer);
      sym_tab [st_index (buffer)].is_output = TRUE;

      if (comma)
      {
         fprintf (out_file, ",\n   %s", buffer);
      }
      else
      {
         fprintf (out_file, "   %s", buffer);
         comma = TRUE;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("$BEGIN statement not found.\n");
         good_bye (TRUE, TRUE);
      }
   }

   if (cnt == 0)
   {
      printf ("No output declared.\n");
      good_bye (TRUE, TRUE);
   }

   fprintf (out_file, " = x;\n\n");


   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $CLOCK statement.\n");
      good_bye (TRUE, TRUE);
   }

   if (strcmp (uppercase (buffer), "$BEGIN"))
   {
      printf ("I cannot translate the $CLOCK statement.\n");
      good_bye (TRUE, TRUE);
   }


   /* reduce the size of symbol table */

   reduce_sym_tab();


   /* patterns */

   curr_time = 0;
   end_of_pat = FALSE;

   while ((fscanf (wsl_file, "%s", buffer) != EOF) && (!end_of_pat))
   {
      if (!is_integer (buffer))
      {
         printf ("Found %s when time was expected. Last time: %lu.\n", 
                                                        buffer, curr_time);
         good_bye (TRUE, TRUE);
      }
      sscanf (buffer, "%ld", &next_time);

      if (next_time != curr_time)
      {
         if (next_time < curr_time)
         {
            printf ("Time cannot proceed backward! Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }

         while ((curr_time = next_period()) < next_time)
         {
            hilo_write_next_period ();
         }
         curr_time = next_time;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
         good_bye (TRUE, TRUE);
      }

      if (buffer[0] != '$')
      {
         us_change (buffer);
         strcpy (element, buffer);

         if (!sym_tab [tmp_index = st_index (uppercase (element))].is_input)
         {
            printf ("%s has not been declared as input.\n", element);
            good_bye (TRUE, TRUE);
         }

         if (sym_tab [tmp_index].last_time == curr_time)
         {
            printf ("Element %s forced more than once at time %lu.\n",
                    element, curr_time);
            good_bye (TRUE, TRUE);
         }

         if (fscanf (wsl_file, "%s", value) == EOF)
         {
            printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }
         if (!is_value (uppercase(value)))
         {
            printf ("Value expected. Last time: %lu.\n", curr_time);
            good_bye (TRUE, TRUE);
         }

         /* filter unnecessary schedulations */
         if (value[0] != sym_tab[tmp_index].last_value)
         {
            fprintf (out_file, "%lu %s = %s;\n", curr_time, element, value);
            sym_tab[tmp_index].last_value = value[0];
            sym_tab[tmp_index].last_time = curr_time;
         }
      }
      else
      {
         if (!strcmp (uppercase (buffer), "$STROBE"))
         {
            fprintf (out_file, "%lu strobe x%lu;\n", curr_time, curr_time);
            last_strobe_time = curr_time;
         }
         else
         if (!strcmp (buffer, "$PERIOD"))
         {
            load_period (curr_time, 'H');
         }
         else
         if (!strcmp (buffer, "$STROBEPERIOD"))
         {
            load_strobe_period (curr_time);
         }
         else
         if (!strcmp (buffer, "$END"))
         {
             end_of_pat = TRUE;
         }
         else
         {
            printf ("%s statement is unknown.\n", buffer);
            good_bye (TRUE, TRUE);
         }
      }
   }


   if (!end_of_pat)
   {
      printf ("Warning: $END statement not found.\n");
   }

   fprintf (out_file, "%lu finish.\n", curr_time);

   return;

}  /* end of hilo_compile */


/******************************************************************************

    Function name:    verilog_compile

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          to compile the wsl file

 ******************************************************************************/


verilog_compile (wave_name)

char  *wave_name;

{

   /* called function */

   unsigned long  next_period();
   char           *get_memory();
   unsigned long  el_index();
   unsigned long  st_index();
   char           *uppercase();
   unsigned char  is_integer();
   unsigned char  is_value();


   /* local variables */

   char           buffer [BUFFER_SIZE];
   char           value [BUFFER_SIZE];
   char           element [BUFFER_SIZE];
   unsigned long  i;
   unsigned long  cnt;
   unsigned long  curr_time;
   unsigned long  next_time;
   unsigned long  tmp_index;
   unsigned char  comma;
   unsigned char  end_of_pat;


   /* beginning of executable part */

   /* write the waveform name */

   fprintf (out_file, "// waveform %spat\n\n", wave_name);

   /* command to trace outputs and inputs */

   fprintf (out_file, "\ninitial\n   $monitor ($time,,\n");
 
   /* read the header */

   /* inputs (list may NOT be empty) */

   fscanf (wsl_file, "%s", buffer);

   if (strcmp (uppercase (buffer), "$INPUT"))
   {
      printf ("I can't find the $INPUT statement.\n");                    
      good_bye (TRUE, TRUE);
   }

   fscanf (wsl_file, "%s", buffer);

   while (strcmp (uppercase (buffer), "$OUTPUT"))
   {
      cnt++;
      sym_tab [st_index (buffer)].is_input = TRUE;

      if (comma)
      {
         fprintf (out_file, ",,\n             %s", buffer);
      }
      else
      {
         fprintf (out_file, "             %s", buffer);
         comma = TRUE;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("I can't find the $OUTPUT statement.\n");
         good_bye (TRUE, TRUE);
      }
   }

   if (cnt == 0)
   {
      printf ("No input declared.\n");
      good_bye (TRUE, TRUE);
   }


   /* outputs */

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $OUTPUT statement.\n");
      good_bye (TRUE, TRUE);
   }

   cnt = 0;

   while (strcmp (uppercase (buffer), "$CLOCK"))
   {
      cnt++;
      sym_tab [st_index (buffer)].is_output = TRUE;

      if (comma)
      {
         fprintf (out_file, ",,\n             %s", buffer);
      }
      else
      {
         fprintf (out_file, "             %s", buffer);
         comma = TRUE;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("$BEGIN statement not found.\n");
         good_bye (TRUE, TRUE);
      }
   }

   if (cnt == 0)
   {
      printf ("No output declared.\n");
      good_bye (TRUE, TRUE);
   }

   fprintf (out_file, ");\n\n");


   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of file after $CLOCK statement.\n");
      good_bye (TRUE, TRUE);
   }

   if (strcmp (uppercase (buffer), "$BEGIN"))
   {
      printf ("I cannot translate the $CLOCK statement.\n");
      good_bye (TRUE, TRUE);
   }


   /* reduce the size of symbol table */

   reduce_sym_tab();


   /* patterns */

   fprintf (out_file, "initial\n   begin\n");

   curr_time = 0;
   end_of_pat = FALSE;

   while ((fscanf (wsl_file, "%s", buffer) != EOF) && (!end_of_pat))
   {
      if (!is_integer (buffer))
      {
         printf ("Found %s when time was expected. Last time: %lu.\n", 
                                                        buffer, curr_time);
         good_bye (TRUE, TRUE);
      }
      sscanf (buffer, "%ld", &next_time);

      if (next_time != curr_time)
      {
         if (next_time < curr_time)
         {
            printf ("Time cannot proceed backward! Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }

         while ((curr_time = next_period()) < next_time)
         {
            verilog_write_next_period ();
         }
         curr_time = next_time;
      }

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
         good_bye (TRUE, TRUE);
      }

      if (buffer[0] != '$')
      {
         strcpy (element, buffer);

         if (!sym_tab [tmp_index = st_index (uppercase (element))].is_input)
         {
            printf ("%s has not been declared as input.\n", element);
            good_bye (TRUE, TRUE);
         }

         if (sym_tab [tmp_index].last_time == curr_time)
         {
            printf ("Element %s forced more than once at time %lu.\n",
                    element, curr_time);
            good_bye (TRUE, TRUE);
         }

         if (fscanf (wsl_file, "%s", value) == EOF)
         {
            printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                    curr_time);
            good_bye (TRUE, TRUE);
         }
         if (!is_value (uppercase(value)))
         {
            printf ("Value expected. Last time: %lu.\n", curr_time);
            good_bye (TRUE, TRUE);
         }

         /* filter unnecessary schedulations */
         if (value[0] != sym_tab[tmp_index].last_value)
         {
            fprintf (out_file, 
                     "      #%lu %s = %s;\n", 
                     curr_time - v_last_time, element, value);
            sym_tab[tmp_index].last_value = value[0];
            sym_tab[tmp_index].last_time = v_last_time = curr_time;
         }
      }
      else
      {
         if (!strcmp (uppercase (buffer), "$STROBE"))
         {
#if 0
            fprintf (out_file, "%lu strobe x%lu;\n", curr_time, curr_time);
#endif
            last_strobe_time = curr_time;
         }
         else
         if (!strcmp (buffer, "$PERIOD"))
         {
            load_period (curr_time, 'H');
         }
         else
         if (!strcmp (buffer, "$STROBEPERIOD"))
         {
            load_strobe_period (curr_time);
         }
         else
         if (!strcmp (buffer, "$END"))
         {
             end_of_pat = TRUE;
         }
         else
         {
            printf ("%s statement is unknown.\n", buffer);
            good_bye (TRUE, TRUE);
         }
      }
   }


   if (!end_of_pat)
   {
      printf ("Warning: $END statement not found.\n");
   }

   fprintf (out_file, "      #%lu $finish;\n", 
            curr_time - v_last_time);
   fprintf (out_file, "\n   end\n\n");

   return;

}  /* end of verilog_compile */


/******************************************************************************

    Function name:    load_period

    Authors:          Pier Luca Montessoro

    Creation date:    6 September 1988

    Purpose:          to load a periodical signal

 ******************************************************************************/


load_period (declare_time, mode)

unsigned long  declare_time;
char           mode;

{

   /* called functions */

   char             *get_memory();
   unsigned long    el_index();
   unsigned long    st_index();
   char             *uppercase();
   unsigned char    is_integer();
   unsigned char    is_value();


   /* local variables */

   char             buffer [BUFFER_SIZE];
   unsigned long    last_time;
   unsigned long    t_index;
   unsigned long    next_time;
   unsigned long    period_next_time;
   unsigned char    end_of_period;
   PERIODIC_SIGNAL  *p_list_ptr;


   /* beginning of executable part */

   /* create the signal descriptor */

   p_list_ptr = (PERIODIC_SIGNAL *) 
                   get_memory ((unsigned long) 1, 
                               (unsigned long) sizeof (PERIODIC_SIGNAL));
   
   p_list_ptr->next_signal = p_list_head;
   p_list_head = p_list_ptr;


   /* load the name of the input */

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                declare_time);
      good_bye (TRUE, TRUE);
   }

   if (mode == 'H')
   {
      us_change (buffer);
   }

   p_list_ptr->st_index = st_index (uppercase(buffer));

   if (mode == 'H')
   {
      if (!sym_tab [p_list_ptr->st_index].is_input)
      {
         printf 
        ("%s has not been declared as input. Period declaration at time %lu.\n",
                             sym_tab [p_list_ptr->st_index].name, declare_time);
         good_bye (TRUE, TRUE);
      }
   }

   p_list_ptr->time_offset = declare_time;
   p_list_ptr->next_value = &(p_list_ptr->table[0]);


   /* load the first transition */

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                declare_time);
      good_bye (TRUE, TRUE);
   }

   if (!is_integer (buffer))
   {
      printf 
         ("Found %s when time was expected. Period declaration at time %lu.\n",
                                                          buffer, declare_time);
      good_bye (TRUE, TRUE);
   }
   sscanf (buffer, "%ld", &next_time);

   if (next_time != 0)
   {
      printf 
         ("First activation of a periodic signal must be at time offset 0.\n");
      printf ("Period declaration at time %lu.\n", declare_time);
      good_bye (TRUE, TRUE);
   }

   p_list_ptr->table[0].time = last_time = 0;
   
   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf 
    ("Premature end of WSL file encountered. Period declaration at time %lu.\n",
                                                                  declare_time);
      good_bye (TRUE, TRUE);
   }
   if (!is_value (uppercase(buffer)))
   {
      printf 
         ("Value expected. Period declaration at time %lu.\n", declare_time);
      good_bye (TRUE, TRUE);
   }

   p_list_ptr->table[0].value = buffer[0];

   t_index = 1;

   end_of_period = FALSE;

   while (!end_of_period)
   {
      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf 
    ("Premature end of WSL file encountered. Period declaration at time %lu.\n",
                                                                  declare_time);
         good_bye (TRUE, TRUE);
      }
      if (!is_integer (buffer))
      {
         printf 
          ("Found %s when time was expected. Period declaration at time %lu.\n",
                                                          buffer, declare_time);
         good_bye (TRUE, TRUE);
      }
      sscanf (buffer, "%ld", &next_time);

      if (next_time <= last_time)
      {
         printf 
            ("Time cannot proceed backward! Period declaration at time %lu.\n",
                                                                  declare_time);
         good_bye (TRUE, TRUE);
      }

      p_list_ptr->table[t_index].time = last_time = next_time;

      if (fscanf (wsl_file, "%s", buffer) == EOF)
      {
         printf 
    ("Premature end of WSL file encountered. Period declaration at time %lu.\n",
                                                                  declare_time);
         good_bye (TRUE, TRUE);
      }

      if ((buffer[0] == '$') && (!strcmp (uppercase (buffer), "$ENDPERIOD")))
      {
         p_list_ptr->table[t_index].value = EOF;
         end_of_period = TRUE;
      }
      else
      {
         if (!is_value (uppercase(buffer)))
         {
            printf ("Value expected. Period declaration at time %lu.\n",
                                                                  declare_time);
            good_bye (TRUE, TRUE);
         }

         p_list_ptr->table[t_index].value = buffer[0];

         if (p_list_ptr->table[t_index].value == 
                             p_list_ptr->table[t_index - 1].value)
         {
            printf ("Warning: repeated schedulations in a period.\n");
         }

         t_index++;
      }
   }

   return;

}  /* end of load_period */


/******************************************************************************

    Function name:    next_period

    Authors:          Pier Luca Montessoro

    Creation date:    6 September 1988

    Purpose:          to search the next activation of a periodical signal

 ******************************************************************************/


unsigned long  next_period ()

{

   /* local variables */

   PERIODIC_SIGNAL  *p_list_ptr;
   unsigned long    next_time;


   /* beginning of executable part */

   p_list_ptr = p_list_head;
   next_time = MAX_UNSIGNED;

   while (p_list_ptr != NULL)
   {
      if (p_list_ptr->next_value->time + p_list_ptr->time_offset < next_time)
      {
         next_time = p_list_ptr->next_value->time + p_list_ptr->time_offset;
         next_p_act_signal = p_list_ptr;
      }
      p_list_ptr = p_list_ptr->next_signal;
   }

   return (next_time);

}  /* end of next_period */


/******************************************************************************

    Function name:    doc_write_next_period

    Authors:          Pier Luca Montessoro

    Creation date:    7 September 1988

    Purpose:          to write the next periodic signal activation in 
                      Dr. Creator format

 ******************************************************************************/


doc_write_next_period ()

{

   /* local variables */

   unsigned long  tmp_index;
   unsigned long  act_time;


   /* beginning of executable part */

   /* check for strobe */

   if (next_p_act_signal->st_index == MAX_UNSIGNED)
   {
      /* strobe signal */

      /* filter multiple activations */

      if (last_strobe_time != next_p_act_signal->time_offset)
      {
         fprintf (out_file, "%d\n", WSL_STROBE);
         last_strobe_time = next_p_act_signal->time_offset;
      }
      next_p_act_signal->time_offset = next_p_act_signal->time_offset + 
                                             next_p_act_signal->table[1].time;

      return;
   }


   /* other signals */

   if (next_p_act_signal->next_value->value == EOF)
   {
      next_p_act_signal->time_offset = 
          next_p_act_signal->time_offset + next_p_act_signal->next_value->time;
      next_p_act_signal->next_value = &(next_p_act_signal->table[0]);
   }

   act_time = next_p_act_signal->time_offset + 
                 next_p_act_signal->next_value->time;

   if (sym_tab[ tmp_index = next_p_act_signal->st_index].last_time == act_time)
   {
      printf ("Element %s forced more than once at time %lu.\n",
               sym_tab[tmp_index].name, act_time);
      good_bye (TRUE, TRUE);
   }

   /* filter unnecessary schedulations */
   if (next_p_act_signal->next_value->value != sym_tab[tmp_index].last_value)
   {
      if (sym_tab[tmp_index].last_time == MAX_UNSIGNED)
      {
         fprintf (out_file, "%d ", WSL_FORCE);
      }
      else
      {
         fprintf (out_file, "%d ", WSL_MOD_FORCE);
      }

      sym_tab[tmp_index].last_time = act_time;
      fprintf (out_file, "%lu %c\n", sym_tab[tmp_index].index, 
                                  next_p_act_signal->next_value->value);

      sym_tab[tmp_index].last_value = next_p_act_signal->next_value->value;
   }

   next_p_act_signal->next_value++;
   next_p_act_signal = NULL;

   return;

}  /* end of doc_write_next_period */


/******************************************************************************

    Function name:    mozart_write_next_period

    Authors:          Pier Luca Montessoro

    Creation date:    7 September 1988

    Purpose:          to write the next periodic signal activation in 
                      Mozart format

 ******************************************************************************/


mozart_write_next_period ()

{

   /* local variables */

   unsigned long  tmp_index;
   unsigned long  act_time;


   /* beginning of executable part */

   if (next_p_act_signal->next_value->value == EOF)
   {
      next_p_act_signal->time_offset = 
          next_p_act_signal->time_offset + next_p_act_signal->next_value->time;
      next_p_act_signal->next_value = &(next_p_act_signal->table[0]);
   }

   act_time = next_p_act_signal->time_offset + 
                 next_p_act_signal->next_value->time;

   if (sym_tab[tmp_index = next_p_act_signal->st_index].last_time == act_time)
   {
      printf ("Element %s forced more than once at time %lu.\n",
               sym_tab[tmp_index].name, act_time);
      good_bye (TRUE, TRUE);
   }

   /* filter unnecessary schedulations */
   if (next_p_act_signal->next_value->value != sym_tab[tmp_index].last_value)
   {
      sym_tab[tmp_index].last_time = act_time;
      fprintf (out_file, "#%lu : %s := %c;\n", act_time, 
               sym_tab[tmp_index].name, next_p_act_signal->next_value->value);
      sym_tab[tmp_index].last_value = next_p_act_signal->next_value->value;
   }

   next_p_act_signal->next_value++;
   next_p_act_signal = NULL;

   return;

}  /* end of mozart_write_next_period */


/******************************************************************************

    Function name:    hilo_write_next_period

    Authors:          Pier Luca Montessoro

    Creation date:    7 September 1988

    Purpose:          to write the next periodic signal activation in 
                      Hilo format

 ******************************************************************************/


hilo_write_next_period ()

{

   /* local variables */

   unsigned long  tmp_index;
   unsigned long  act_time;


   /* beginning of executable part */

   /* check for strobe */

   if (next_p_act_signal->st_index == MAX_UNSIGNED)
   {
      /* strobe signal */

      /* filter multiple activations */

      if (last_strobe_time != next_p_act_signal->time_offset)
      {
         fprintf (out_file, "%lu strobe x%lu;\n", 
                            next_p_act_signal->time_offset, 
                            next_p_act_signal->time_offset);
      }
      last_strobe_time = next_p_act_signal->time_offset ;
      next_p_act_signal->time_offset = next_p_act_signal->time_offset + 
                                           next_p_act_signal->table[1].time;

      return;
   }


   /* other signals */

   if (next_p_act_signal->next_value->value == EOF)
   {
      next_p_act_signal->time_offset = 
          next_p_act_signal->time_offset + next_p_act_signal->next_value->time;
      next_p_act_signal->next_value = &(next_p_act_signal->table[0]);
   }

   act_time = next_p_act_signal->time_offset + 
                 next_p_act_signal->next_value->time;

   if (sym_tab[tmp_index = next_p_act_signal->st_index].last_time == act_time)
   {
      printf ("Element %s forced more than once at time %lu.\n",
               sym_tab[tmp_index].name, act_time);
      good_bye (TRUE, TRUE);
   }

   /* filter unnecessary schedulations */
   if (next_p_act_signal->next_value->value != sym_tab[tmp_index].last_value)
   {
      sym_tab[tmp_index].last_time = act_time;
      fprintf (out_file, "%lu %s = %c;\n", act_time, sym_tab[tmp_index].name, 
                                          next_p_act_signal->next_value->value);
      sym_tab[tmp_index].last_value = next_p_act_signal->next_value->value;
   }

   next_p_act_signal->next_value++;
   next_p_act_signal = NULL;

   return;

}  /* end of hilo_write_next_period */


/******************************************************************************

    Function name:    verilog_write_next_period

    Authors:          Pier Luca Montessoro

    Creation date:    7 September 1988

    Purpose:          to write the next periodic signal activation in 
                      Verilog format

 ******************************************************************************/


verilog_write_next_period ()

{

   /* local variables */

   unsigned long  tmp_index;
   unsigned long  act_time;


   /* beginning of executable part */

   /* check for strobe */

   if (next_p_act_signal->st_index == MAX_UNSIGNED)
   {
      /* strobe signal */

      /* filter multiple activations */

      if (last_strobe_time != next_p_act_signal->time_offset)
      {
#if 0
         fprintf (out_file, "%lu strobe x%lu;\n", 
                            next_p_act_signal->time_offset, 
                            next_p_act_signal->time_offset);
#endif
      }
      last_strobe_time = next_p_act_signal->time_offset ;
      next_p_act_signal->time_offset = next_p_act_signal->time_offset + 
                                           next_p_act_signal->table[1].time;

      return;
   }


   /* other signals */

   if (next_p_act_signal->next_value->value == EOF)
   {
      next_p_act_signal->time_offset = 
          next_p_act_signal->time_offset + next_p_act_signal->next_value->time;
      next_p_act_signal->next_value = &(next_p_act_signal->table[0]);
   }

   act_time = next_p_act_signal->time_offset + 
                 next_p_act_signal->next_value->time;

   if (sym_tab[tmp_index = next_p_act_signal->st_index].last_time == act_time)
   {
      printf ("Element %s forced more than once at time %lu.\n",
               sym_tab[tmp_index].name, act_time);
      good_bye (TRUE, TRUE);
   }

   /* filter unnecessary schedulations */
   if (next_p_act_signal->next_value->value != sym_tab[tmp_index].last_value)
   {
      fprintf (out_file, "      #%lu %s = %c;\n", 
               act_time - v_last_time, 
               sym_tab[tmp_index].name, 
               next_p_act_signal->next_value->value);
      sym_tab[tmp_index].last_time = v_last_time = act_time;
      sym_tab[tmp_index].last_value = next_p_act_signal->next_value->value;
   }

   next_p_act_signal->next_value++;
   next_p_act_signal = NULL;

   return;

}  /* end of verilog_write_next_period */


/******************************************************************************

    Function name:    load_strobe_period

    Authors:          Pier Luca Montessoro

    Creation date:    6 September 1988

    Purpose:          to load a periodical strobe signal

 ******************************************************************************/


load_strobe_period (declare_time)

unsigned long  declare_time;

{

   /* called functions */

   char             *get_memory();
   unsigned long    el_index();
   unsigned long    st_index();
   char             *uppercase();
   unsigned char    is_integer();
   unsigned char    is_value();


   /* local variables */

   char             buffer [BUFFER_SIZE];
   unsigned long    last_time;
   unsigned long    t_index;
   unsigned long    next_time;
   unsigned long    period_next_time;
   unsigned char    end_of_period;
   PERIODIC_SIGNAL  *p_list_ptr;


   /* beginning of executable part */

   /* create the signal descriptor */

   p_list_ptr = (PERIODIC_SIGNAL *) 
                   get_memory ((unsigned long) 1, 
                               (unsigned long) sizeof (PERIODIC_SIGNAL));
   
   p_list_ptr->next_signal = p_list_head;
   p_list_head = p_list_ptr;

   /* mark that the signal is a strobe */

   p_list_ptr->st_index = MAX_UNSIGNED;

   p_list_ptr->time_offset = declare_time;
   p_list_ptr->next_value = &(p_list_ptr->table[0]);


   /* load the time period */

   if (fscanf (wsl_file, "%s", buffer) == EOF)
   {
      printf ("Premature end of WSL file encountered. Last time: %lu.\n", 
                                                                declare_time);
      good_bye (TRUE, TRUE);
   }

   if (!is_integer (buffer))
   {
      printf 
         ("Found %s when time was expected. Period declaration at time %lu.\n",
                                                          buffer, declare_time);
   }
   sscanf (buffer, "%ld", &next_time);

   if (next_time == 0)
   {
      printf 
         ("Strobe period cannot be 0. Period declaration at time %lu.\n",
                                                                declare_time);
      good_bye (TRUE, TRUE);
   }

   p_list_ptr->table[0].time = 0;
   p_list_ptr->table[0].value = 'X';

   p_list_ptr->table[1].time = next_time;
   p_list_ptr->table[1].value = EOF;

   return;

}  /* end of load_strobe_period */


/******************************************************************************

    Function name:    good_bye

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          exit function

 ******************************************************************************/


good_bye (open_files, error_found)

unsigned char  open_files;
unsigned char  error_found;

{

   /* beginning of executable part */

   if (open_files)
   {
      fclose (wsl_file);
      fclose (out_file);
   }

   if (error_found)
   {
      printf ("I am sorry, but I cannot proceed. Try again.\n");
   }
   else
   {
      printf ("Job terminated. Good-bye!");
   }

   exit (1);

}  /* end of good_bye */


/******************************************************************************

    Function name:    get_memory

    Authors:          Pier Luca Montessoro

    Creation date:    25 January 1988

    Purpose:          memory allocation for network data structures

 ******************************************************************************/


char  *get_memory (number, size)

unsigned long  number;
unsigned long  size;

{

   /* local variable */

   char   *alloc_pointer;


   /* beginning of executable part */

   alloc_pointer = (char *) malloc (number * size);


   if (alloc_pointer == NULL)
   {
      printf ("No memory is available for topological data allocation.\n");
      good_bye (TRUE, TRUE);
   }

   return (alloc_pointer);

}  /* end of get_memory */


/******************************************************************************

    Function name:    load_sym_tab

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to load the symbol table of the current experiment

 ******************************************************************************/


load_sym_tab (exp_name, mode)

char  *exp_name;
char  mode;

{

   /* local variables */

   char                  file_name [MAX_FILE_NAME_LENGTH];
   FILE                  *sym_tab_file;
   char                  *el_name;
   char                  dummy_str [MAX_NAME_LENGTH];
   unsigned long         dummy_num;
   unsigned long         mod_size;
   unsigned long         n_modules;
   unsigned long         it_size;
   unsigned long         t_index;
   unsigned long         i;
   unsigned long         j;


   /* beginning of executable part */

   /* open the file containing the symbol table */

   sprintf (file_name, "%s.%s", exp_name, SYMTAB_EXTENSION);

   if ((sym_tab_file = fopen (file_name, "r")) == NULL)
   {
      printf ("Symbol table file not found.\n");
      good_bye (TRUE, TRUE);
   }


   /* read the number of modules */
   fscanf (sym_tab_file, "%ld %ld", &n_modules, &dummy_num);


   /* reach the main module (the first one) */

   for (j = 0; j < n_modules; j++)
   {

      /* read the size of the interface table */   
      fscanf (sym_tab_file, "%s %ld %ld %s %ld %ld %ld %ld", 
              dummy_str, &dummy_num, &dummy_num, dummy_str, 
              &dummy_num, &dummy_num, &dummy_num, 
              &it_size);

      /* skip the interface table (it should be empty for the main module) */
      for (i = 0; i < it_size; i++)
      {
         fscanf (sym_tab_file, "%ld %s", &dummy_num, dummy_str);
      }
   }


   /* read the module name and the size */
   fscanf (sym_tab_file, "%s %ld", dummy_str, &mod_size);

   /* alloc the memory needed for the symbol table */
   if ((sym_tab = (SYM_TAB *) 
             get_memory (mod_size, (unsigned long) sizeof (SYM_TAB))) == NULL)
   {
      printf ("I cannot alloc memory for the symbol table.\n");
      fclose (sym_tab_file);
      good_bye (TRUE, TRUE);
   }

   sym_tab_dim = mod_size;

   /* load the names */

   us_converted = FALSE;

   for (i = 0; i < mod_size; i++)
   {
      fscanf (sym_tab_file, "%ld %ld %s", &t_index, &dummy_num, dummy_str);

      sym_tab[i].index = t_index;
      sym_tab[i].is_output = FALSE;
      sym_tab[i].is_input = FALSE;
      sym_tab[i].last_value = 'X';
      sym_tab[i].last_time = MAX_UNSIGNED;

      /* convert the name to uppercase */
      el_name = sym_tab[i].name;
      if ((mode == 'h') || (mode == 'H'))
      {
         for (j = 0; dummy_str[j] != '\0'; j++)
         {
            if (dummy_str[j] == '_')
            {
               us_converted = TRUE;
               el_name[j] = 'X';
            }
            else
            {
               el_name[j] = toupper (dummy_str[j]);
            }
         }
      }
      else
      {
         for (j = 0; dummy_str[j] != '\0'; j++)
         {
            el_name[j] = toupper (dummy_str[j]);
         }
      }
      el_name[j] = '\0';
   }

   fclose (sym_tab_file);

   printf ("Symbol table loaded.\n");

   if (us_converted)
   {
      printf ("Warning: some '_' in names have been converted to 'X'.\n");
   }


   return;

}  /* end of load_sym_tab */


/******************************************************************************

    Function name:    reduce_sym_tab

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to remove useless entries in symbol table

 ******************************************************************************/


reduce_sym_tab ()

{

   /* local variables */

   unsigned long         i;
   unsigned long         j;


   /* beginning of executable part */

   for (i = 0, j = 0; i < sym_tab_dim; i++)
   {
      if (sym_tab[i].is_output || sym_tab[i].is_input || sym_tab[i].is_clock)
      {
         sym_tab[j].index = sym_tab[i].index;
         sym_tab[j].is_output = sym_tab[i].is_output;
         sym_tab[j].is_input = sym_tab[i].is_input;
         sym_tab[j].last_value = sym_tab[i].last_value;
         sym_tab[j].last_time = sym_tab[i].last_time;
         strcpy (sym_tab[j].name, sym_tab[i].name);
         j++;
      }
   }

   sym_tab_dim = j;

   return;

}  /* end of reduce_sym_tab */


/******************************************************************************

    Function name:    el_index

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to find the index of an element

 ******************************************************************************/


unsigned long  el_index (name)

char  *name;

{

   /* local variables */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; i < sym_tab_dim; i++)
   {
      if (!strcmp (sym_tab[i].name, name))
      {
         return (sym_tab[i].index);
      }
   }


   printf ("I cannot find element %s in the main symbol table.\n", name);

   good_bye (TRUE, TRUE);


}  /* end of el_index */


/******************************************************************************

    Function name:    st_index

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to find the index of an element

 ******************************************************************************/


unsigned long  st_index (name)

char  *name;

{

   /* local variables */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; i < sym_tab_dim; i++)
   {
      if (!strcmp (sym_tab[i].name, name))
      {
         return (i);
      }
   }


   printf ("I cannot find element %s in the main symbol table.\n", name);

   good_bye (TRUE, TRUE);


}  /* end of st_index */


/******************************************************************************

    Function name:    uppercase

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to convert a string to uppercase

 ******************************************************************************/


char *uppercase (name)

char  *name;

{

   /* local variable */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; name[i] != '\0'; name [i] = toupper (name [i]), i++);
   return (name);

}  /* end of uppercase */


/******************************************************************************

    Function name:    us_change

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to convert the '_' in a string into 'X'

 ******************************************************************************/


us_change (str)

char  *str;

{

   /* local variable */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; str[i] != '\0'; i++)
   {
      if (str[i] == '_')
      {
         str[i] = 'X';
      }
   }

   return;

}  /* end of us_change */


/******************************************************************************

    Function name:    is_integer

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to check the type of a string

 ******************************************************************************/


unsigned char is_integer (str)

char  str[];

{

   /* local variable */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; str[i] != '\0'; i++)
   {
      if ((str[i] < '0') || (str[i] > '9'))
      {
         return (FALSE);
      }
   }

   return (TRUE);

}  /* end of is_integer */


/******************************************************************************

    Function name:    is_value

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to check the type of a string

 ******************************************************************************/


unsigned char is_value (str)

char  *str;

{

   if ((str[1] != '\0') || 
       ((str[0] != '0') && (str[0] != '1') && 
        (str[0] != 'X') && (str[0] != 'Z')))
   {
      return (FALSE);
   }
   else
   {
      return (TRUE);
   }

}  /* end of is_value */

